# Volatility
# Copyright (C) 2008 Volatile Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details. 
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
#

"""
@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""

import forensics.win32.gdi as gdi

from forensics.win32.gdi import get_flags
from forensics.object2 import *
from forensics.win32.tasks import process_list
from forensics.win32.lists import list_entry
from struct import unpack
from vutils import *

import pygame

def r2pr(tup):
    return tup[0], tup[1], tup[2]-tup[0], tup[3]-tup[1]

def get_threads(proc):
    return list_entry(proc.vm, types, proc.profile, 
                      proc.ThreadListHead.v(), "_ETHREAD",
                      fieldname="ThreadListEntry")

def draw_text(screen, font, text, left, top, color=0):
    lines = text.split('\x0d\x0a')
    for line in lines:
        surf = font.render(line, True, color)
        screen.blit(surf, (left, top))
        _, _, _, height = surf.get_rect()
        top += height

class wndmon(forensics.commands.command):

    # Declare meta information associated with this plugin
    
    meta_info = forensics.commands.command.meta_info 
    meta_info['author'] = 'Brendan Dolan-Gavitt'
    meta_info['copyright'] = 'Copyright (c) 2007,2008 Brendan Dolan-Gavitt'
    meta_info['contact'] = 'bdolangavitt@wesleyan.edu'
    meta_info['license'] = 'GNU General Public License 2.0 or later'
    meta_info['url'] = 'http://moyix.blogspot.com/'
    meta_info['os'] = 'WIN_32_XP_SP2'
    meta_info['version'] = '1.0'

    def parser(self):
        forensics.commands.command.parser(self)

        self.op.add_option('-a', '--atom-table',
            help='address of atom table',
            type='int', dest='atom')
        
        # Atom table addresses: Xen  0x1a8b98
        #                       NIST 0x1a70f8

    def help(self):
        return  "Show on-screen windows in near-realtime"
    
    def find_top_window(self, process_address_space, types, symtab, theProfile):
        top_windows = []
        windows = set()
        pslist = process_list(process_address_space, types, symtab)
        for p in pslist:
            proc = Object("_EPROCESS", p, process_address_space, profile=theProfile)
            if not proc.vm: continue
            win = None
            for thrd in get_threads(proc):
                win = gdi.get_desktop_window(thrd)
                if win: break

            if win and win not in windows:
                top_windows.append( (win, thrd, proc) )
                windows.add(win)

        return top_windows

    def execute(self):
        from vtypes import xpsp2types
        
        xpsp2types['_ETHREAD'][1]['ThreadListEntry'] = [ 0x22c, ['_LIST_ENTRY']]
        xpsp2types['_KTHREAD'][1]['Win32Thread'] = [ 0x130, ['pointer', ['_W32THREAD']]]
        xpsp2types['_KTHREAD'][1]['ServiceTable'] = [ 0xe0, ['pointer', ['void']]]

        theProfile = Profile()
        theProfile.add_types(gdi.gdi_types)

        (addr_space, symtab, types) = load_and_identify_image(self.op,
            self.opts)

        # Init pygame
        window = pygame.display.set_mode((1024,768))
        pygame.display.set_caption('Guest View') 
        screen = pygame.display.get_surface() 
        pygame.font.init()
        font = pygame.font.Font('/usr/share/fonts/truetype/msttcorefonts/arial.ttf', 12)
        fontb = pygame.font.Font('/usr/share/fonts/truetype/msttcorefonts/arialbd.ttf', 12)
        black = pygame.color.Color("Black")
        white = pygame.color.Color("White")
        
        # Load the atom table
        if self.opts.atom:
            AtomTableAddrOff = self.opts.atom
            win32k_base = gdi.find_win32k(addr_space, types, symtab)
            atom_table_off = win32k_base+AtomTableAddrOff
        else:
            from forensics.win32.findatomtable import find_atomtable
            atom_table_off = find_atomtable(addr_space, types, symtab, theProfile)

        addr_space = gdi.find_space(addr_space, types, symtab, atom_table_off)
        UserAtomTableHandle = read_value(addr_space, 'pointer', atom_table_off)
        atoms = gdi.read_atoms(addr_space, UserAtomTableHandle, theProfile)

        wins = self.find_top_window(addr_space,types,symtab,theProfile)
        wins = wins[:1]
        while True:
            for win, thrd, proc in wins:
                # Update the pygame screen
                # Closure is used here so that we can have the screen instance be the same
                # every time the callback is triggered.
                def draw_win(win,level):
                    screen.fill(pygame.color.Color("White"), r2pr(win.rcWnd.get_tup()))
                    pygame.draw.rect(screen, pygame.color.Color("Black"), r2pr(win.rcWnd.get_tup()), 1)
                    screen.fill(pygame.color.Color("White"), r2pr(win.rcClient.get_tup()))
                    pygame.draw.rect(screen, pygame.color.Color("Black"), r2pr(win.rcClient.get_tup()), 1)

                    text = None
                    if atoms[win.pClassInfo.SuperclassAtom].Name == 'RichEdit20W':
                        text = gdi.get_richedit_text(win)
                    elif atoms[win.pClassInfo.SuperclassAtom].Name == 'Edit':
                        text = gdi.get_editbox_text(win)
                    elif win.WindowText.Buffer:
                        text = win.WindowText.Buffer.encode('ascii', 'backslashreplace')

                    if not text:
                        text = ''

                    draw_text(screen, fontb, text, win.rcWnd.left+2, win.rcWnd.top, color=black)

                    if atoms[win.pClassInfo.SuperclassAtom].Name == "ToolBarWindow32":
                        for name,rect in gdi.get_tb_buttons(win):
                            screen.fill(pygame.color.Color("White"), r2pr(rect))
                            pygame.draw.rect(screen, pygame.color.Color("Black"), r2pr(rect), 1)
                            surf = font.render(name, True, black)
                            screen.blit(surf, (rect[0]+2, rect[1]))

                gdi.traverse_windows(win, fun=draw_win, do_node=gdi.is_visible)
                pygame.display.flip()
